@@ -138,7 +138,9 @@ adwaita_engine_render_focus (GtkThemingEngine *engine,
   gint border_radius;
   gboolean use_dashes;
   double dashes[2] = { 2.0, 0.2 };
+  const GtkWidgetPath *path;
 
+  path = gtk_theming_engine_get_path (engine);
   state = gtk_theming_engine_get_state (engine);
   gtk_theming_engine_get (engine, state,
                           "-adwaita-focus-border-color", &border_color,
@@ -153,16 +155,27 @@ adwaita_engine_render_focus (GtkThemingEngine *engine,
                                 "focus-padding", &focus_pad,
                                 NULL);
 
+  /* as we render the tab smaller than the whole allocation, we need
+   * to recenter and resize the focus on the tab.
+   */
   if (gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_NOTEBOOK) &&
       gtk_theming_engine_has_region (engine, GTK_STYLE_REGION_TAB, NULL))
     {
-      /* as we render the tab smaller than the whole allocation, we need
-       * to recenter and resize the focus on the tab.
-       */
       y += 3.0;
       height -= 3.0;
     }
 
+  /* the treeview rows don't change allocation when modifying focus-padding,
+   * so we have to move the focus ring inside the allocated area manually.
+   */
+  if (gtk_widget_path_is_type (path, GTK_TYPE_TREE_VIEW))
+    {
+      x += focus_pad;
+      y += focus_pad;
+      width -= 2 * focus_pad;
+      height -= 2 * focus_pad;
+    }
+
   cairo_save (cr);
   cairo_set_line_width (cr, line_width);
 
@@ -481,12 +494,30 @@ adwaita_engine_render_expander (GtkThemingEngine *engine,
   gdouble side, offset;
   gint line_width;
   GtkBorder border;
+  const GtkWidgetPath *path = gtk_theming_engine_get_path (engine);
 
   side = floor (MIN (width, height));
 
+  if (gtk_widget_path_is_type (path, GTK_TYPE_TREE_VIEW) &&
+      (side == 17))
+    {
+      /* HACK: draw the expander as if it was 11px instead of the allocated 17px,
+       * so that we can have a bit of padding between the view edge and the
+       * expander itself.
+       */
+      x += 3;
+      y += 3;
+      width -= 6;
+      height -= 6;
+      side -= 6;
+    }
+
   x += width / 2 - side / 2;
   y += height / 2 - side / 2;
 
+  x = floor (x);
+  y = floor (y);
+
   /* make sure the rendered side length is always odd */
   if (((gint) side % 2) == 0)
     side -= 1.0;
@@ -768,50 +799,6 @@ adwaita_engine_render_handle (GtkThemingEngine *engine,
     }
 }
 
-static void
-adwaita_engine_render_line (GtkThemingEngine *engine,
-                            cairo_t *cr,
-                            gdouble x0,
-                            gdouble y0,
-                            gdouble x1,
-                            gdouble y1)
-{
-  const GtkWidgetPath *path;
-
-  path = gtk_theming_engine_get_path (engine);
-
-  if ((gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_MARK) &&
-       gtk_widget_path_is_type (path, GTK_TYPE_SCALE)) ||
-      (gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_SEPARATOR) &&
-       gtk_widget_path_is_type (path, GTK_TYPE_TREE_VIEW)))
-    {
-      GtkStateFlags state;
-      GdkRGBA bg;
-
-      state = gtk_theming_engine_get_state (engine);
-      gtk_theming_engine_get_background_color (engine, state, &bg);
-
-      cairo_save (cr);
-
-      cairo_set_line_cap (cr, CAIRO_LINE_CAP_SQUARE);
-      cairo_set_line_width (cr, 1);
-
-      cairo_move_to (cr, x0 + 0.5, y0 + 0.5);
-      cairo_line_to (cr, x1 + 0.5, y1 + 0.5);
-
-      gdk_cairo_set_source_rgba (cr, &bg);
-      cairo_stroke (cr);
-
-      cairo_restore (cr);
-    }
-  else
-    {
-      GTK_THEMING_ENGINE_CLASS (adwaita_engine_parent_class)->render_line
-        (engine, cr,
-         x0, y0, x1, y1);
-    }
-}
-
 static void
 adwaita_engine_class_init (AdwaitaEngineClass *klass)
 {
@@ -828,7 +815,6 @@ adwaita_engine_class_init (AdwaitaEngineClass *klass)
   engine_class->render_activity = adwaita_engine_render_activity;
   engine_class->render_slider = adwaita_engine_render_slider;
   engine_class->render_handle = adwaita_engine_render_handle;
-  engine_class->render_line = adwaita_engine_render_line;
 
   gtk_theming_engine_register_property (ADWAITA_NAMESPACE, NULL,
                                         g_param_spec_boxed ("focus-border-color",
